#include "Annotation.h"

#include <GL/glut.h>
#include <GL/glu.h>
#include <GL/gl.h>


//-----------------------------------------------------------------------------
Annotation::Annotation()
{
    m_ctrlDown = false;
    m_mouseDown = false;
    m_pen_angle = 0.1;
    m_max_penrange = 10000;
    m_pen_range = m_max_penrange;
    m_show_penrange = 0;

    m_currentLabel = std::make_tuple(1,1,1);
}
//-----------------------------------------------------------------------------
void Annotation::initialize(
        std::vector<std::shared_ptr<PMap::PointCloud3d > >* pointClouds,
        std::vector<bool>*                                  activePointClouds,
        std::function<void(int)>                            activatePointCloud
    )
{
    m_PointClouds = pointClouds;
    m_activePointClouds = activePointClouds;
    f_activatePointCloud = activatePointCloud;
}
//-----------------------------------------------------------------------------
void Annotation::render_pen(QPoint cursor_pos )
{

    float width = tan(m_pen_angle);

    double mdl[16];     // modelviewmatrix
    double proj[16];    // projection matrix
    int viewp[4];       // viewport
    glGetDoublev(GL_MODELVIEW_MATRIX, mdl);
    glGetDoublev(GL_PROJECTION_MATRIX, proj);
    glGetIntegerv(GL_VIEWPORT, viewp);

    glLineWidth(1);

    float alpha = 0.3;
    if(m_ctrlDown)
        alpha = 0.5;

    glColor4f(std::get<0>(m_currentLabel)/1.0,
              std::get<1>(m_currentLabel)/1.0,
              std::get<2>(m_currentLabel)/1.0,
              alpha);




    glEnable(GL_LINE_SMOOTH);

    GLdouble dx, dy, dz;
    gluUnProject(cursor_pos.x(), (float)viewp[3] - cursor_pos.y(), 1, mdl, proj, viewp, &dx, &dy, &dz);

    Eigen::Vector3d origin = m_camera->getPos();       // m -> cm
    Eigen::Vector3d direction = (Eigen::Vector3d(dx, dy, dz))- origin;
    direction.normalize();


    origin += direction;


    // render Pen-Sphere

    glEnable(GL_BLEND);
    glEnable(GL_DEPTH_TEST);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glPushMatrix();

    glTranslated(origin(0),origin(1),origin(2));
    glutSolidSphere(width,30,30);
    glPopMatrix();

    glDisable(GL_BLEND);
    glDisable(GL_DEPTH_TEST);


    // render Pen-Circle

    int n = 20;

    Eigen::Vector3d h(1,0,0);
    Eigen::Vector3d s(0,1,0);

    if((direction.x() != 0)||(direction.y() != 0))
    {
        h(0) = direction.y();
        h(1) = direction.x() * -1;
        h(2) = 0;
        if(direction.z() == 0)
        {
            s(0) = 0;
            s(1) = 0;
            s(2) = 1;
        }
        else
        {
            Eigen::Vector3d dxy = direction;
            dxy(2) = 0;
            s = -dxy;
            s.normalize();
            s(2) = dxy.norm() / direction.z();
        }
    }
    h.normalize();
    s.normalize();
    h *= width;
    s *= width;

    glBegin(GL_LINE_STRIP);

    for(int i=0; i <= n; i++)
    {
        Eigen::Vector3d p = origin  + sin(i * 2 * M_PI / n) * h
                + cos(i*2*M_PI/n) * s;
        glVertex3f(p.x(), p.y(), p.z());

    }

    glEnd();
    glDisable(GL_LINE_SMOOTH);


    // render pen-Range
    if(m_show_penrange)
    {
        m_show_penrange--;
        glColor3f(0,0,1);

        char buf[100];
        if(m_pen_range == m_max_penrange)
            sprintf(buf,"         pen range: infinite");
        else
            sprintf(buf,"         pen range: %.0f m", m_pen_range/100);

        const char * t = buf;

        glRasterPos3f(origin.x(), origin.y(), origin.z());
        do glutBitmapCharacter( GLUT_BITMAP_HELVETICA_18, *t ); while( *(++t) ) ;
    }
}
//-----------------------------------------------------------------------------
void Annotation::draw(QPointF mouseXY)
{

    // -------- calculate Line Of Sight ---------------

    double mdl[16];     // modelviewmatrix
    double proj[16];    // projection matrix
    int viewp[4];       // viewport
    glGetDoublev(GL_MODELVIEW_MATRIX, mdl);
    glGetDoublev(GL_PROJECTION_MATRIX, proj);
    glGetIntegerv(GL_VIEWPORT, viewp);

    double cos_pen_angle = cos(m_pen_angle);

    // ray direction (d)
    GLdouble dx, dy, dz;
    gluUnProject(mouseXY.x(), (float)viewp[3] - mouseXY.y(), 1, mdl, proj, viewp, &dx, &dy, &dz);

    Eigen::Vector3d  origin = m_camera->getPos() * 100;       // m -> cm
    Eigen::Vector3d  direction = (Eigen::Vector3d(dx, dy, dz) * 100 ) - origin;
    direction.normalize();

    // --------- search near points ------------------

    for(uint n_pntCloud = 0; n_pntCloud < m_PointClouds->size(); n_pntCloud++)
        if(m_activePointClouds->at(n_pntCloud))
            for(uint n_pnt = 0; n_pnt < m_PointClouds->at(n_pntCloud)->getNumPoints(); n_pnt++)
            {
                Eigen::Vector3d cam2pnt =  (m_PointClouds->at(n_pntCloud)->getPoints().col(n_pnt) - origin );

                double dot = direction.dot(cam2pnt);
                if((dot < 0) ||((dot >= m_pen_range)&&(m_pen_range < m_max_penrange)) )  // behind camera
                    continue;

                double distance_point_origin = cam2pnt.norm();

                if(cos_pen_angle <= dot/distance_point_origin)
                    m_PointClouds->at(n_pntCloud)->setColor(n_pnt, m_currentLabel);

            }
}
//-----------------------------------------------------------------------------
void Annotation::checkForPCActivation(QPointF mouseXY)
{


    // -------- calculate Line Of Sight ---------------

    double mdl[16];     // modelviewmatrix
    double proj[16];    // projection matrix
    int viewp[4];       // viewport
    glGetDoublev(GL_MODELVIEW_MATRIX, mdl);
    glGetDoublev(GL_PROJECTION_MATRIX, proj);
    glGetIntegerv(GL_VIEWPORT, viewp);

    // ray direction (d)
    GLdouble dx, dy, dz;
    gluUnProject(mouseXY.x(), (float)viewp[3] - mouseXY.y(), 1, mdl, proj, viewp, &dx, &dy, &dz);

    Eigen::Vector3d origin = m_camera->getPos();       // m -> cm
    Eigen::Vector3d direction = (Eigen::Vector3d(dx, dy, dz)  ) - origin;
    direction.normalize();

    // --------- search near centers ------------------

    for(uint n_pntCloud = 0; n_pntCloud < m_PointClouds->size(); n_pntCloud++)
    {
        Eigen::Vector3d cam2pnt =  ((m_PointClouds->at(n_pntCloud)->getOrigin().cast<double>()) - origin );
        if(direction.dot(cam2pnt) < 0)  // behind camera
            continue;

        double distance_point_LOS = (cam2pnt.cross(direction)).norm();

        if(distance_point_LOS <= 0.5)  // inside pen-cone
            f_activatePointCloud(n_pntCloud);
    }

}
//-----------------------------------------------------------------------------
void Annotation::keyPressEvent(QKeyEvent *e)
{
    if(e->key()==Qt::Key_Control)
        m_ctrlDown = true;
    else if(e->key()==Qt::Key_Shift)
        m_shiftDown = true;

}
//-----------------------------------------------------------------------------
void Annotation::keyReleaseEvent(QKeyEvent *e)
{

    if(e->key()==Qt::Key_Control)
        m_ctrlDown = false;
    else if(e->key()==Qt::Key_Shift)
        m_shiftDown = false;
}
//-----------------------------------------------------------------------------
void Annotation::mousePressEvent(QMouseEvent* e)
{
    m_mouseDown = true;
    if(m_ctrlDown)
        draw(e->posF());
}
//-----------------------------------------------------------------------------
void Annotation::mouseMoveEvent(QMouseEvent *e)
{
    if((m_ctrlDown)&&(m_mouseDown))
        draw(e->posF());
}
//-----------------------------------------------------------------------------
void Annotation::mouseReleaseEvent(QMouseEvent* e)
{
    m_mouseDown = false;
}
//-----------------------------------------------------------------------------
void Annotation::wheelEvent(QWheelEvent* e)
{

    if(m_shiftDown)
    {
        if(e->delta() > 0)
        {
            if(m_pen_range < m_max_penrange)
                m_pen_range += 100;
        }
        else
            if(m_pen_range > 100)
                m_pen_range -= 100;

        m_show_penrange = 50;
    }
    else
    {
        if(e->delta() > 0)
            m_pen_angle += 0.01;
        else
            if(m_pen_angle > 0.01)
                m_pen_angle -= 0.01;
    }
}
//-----------------------------------------------------------------------------
void Annotation::mouseDoubleClickEvent(QMouseEvent* e)
{
    checkForPCActivation(e->posF());
}
//-----------------------------------------------------------------------------
