#include "MainWindow.h"
#include "ui_MainWindow.h"
#include "CVUtil.h"
#include "Timing.h"
#include "Annotation.h"

#include <iostream>
#include <fstream>
#include <boost/foreach.hpp>
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/json_parser.hpp>

#include <GL/glut.h>


//-----------------------------------------------------------------------------
MainWindow::MainWindow(QWidget *parent) : QMainWindow{parent}, m_ui{new Ui::MainWindow}, m_renderHeightColors(false)
{
    m_ui->setupUi(this);

    m_ui->scanTreeWidget->setSelectionBehavior(QAbstractItemView::SelectRows);
    m_ui->scanTreeWidget->setColumnCount(2);
    m_ui->scanTreeWidget->setHeaderLabels(QStringList() << "Dataset" << "Path");
    m_ui->scanTreeWidget->header()->resizeSection(0 /*column index*/, 110 /*width*/);
    m_ui->scanTreeWidget->sortByColumn(0, Qt::AscendingOrder);
    m_ui->scanTreeWidget->setSortingEnabled(true);

    connect(m_ui->loadButton, SIGNAL(clicked()), this, SLOT(loadDataset()));
    connect(m_ui->saveButton, SIGNAL(clicked()), this, SLOT(saveDataset()));
    connect(m_ui->pointSizeBox, SIGNAL(valueChanged(int)), this, SLOT(pointSizeChanged(int)));
    connect(m_ui->scanTreeWidget, SIGNAL(itemSelectionChanged()), this, SLOT(loadVelodata()));
    connect(m_ui->labelBox, SIGNAL(currentIndexChanged(QString)), this, SLOT(labelChanged(QString)));
    connect(m_ui->showPointsCheckBox, SIGNAL(stateChanged(int)), this, SLOT(showPointsChanged(int)));

    m_Annotation.initialize(&m_PointClouds, &m_ActivePointClouds, [&](int i){activatePointCloud(i);});
    m_ui->widget->setAnnotation(&m_Annotation);

}
//-----------------------------------------------------------------------------
MainWindow::~MainWindow()
{
}
//-----------------------------------------------------------------------------
void MainWindow::loadVelodata()
{
    // clear old Pointclouds
    m_ui->widget->clearRenderables();
    m_cloudRenderer.clear(); //smart pointers are also deleted
    for(unsigned int i = 0; i< m_ActivePointClouds.size(); i++)
        m_ActivePointClouds.at(i) = false;

    // load new Pointclouds
    for(int i = 0; i < m_ui->scanTreeWidget->selectedItems().size(); i++) //iterate over all point cloud selections
    {
        const QTreeWidgetItem* item = m_ui->scanTreeWidget->selectedItems().at(i);
        for(unsigned int j = 0; j < m_PointClouds.size(); j++)  //iterate over all point clouds
        {
            if(QFileInfo(item->data(1,Qt::DisplayRole).toString()).baseName() == QFileInfo(QString::fromStdString(m_PointClouds.at(j)->getLabelFile())).baseName())
            { //only the baseName() function always yields a unique identifier
                m_cloudRenderer.emplace_back(new PointCloudRenderer(m_PointClouds.at(j)));
                m_cloudRenderer.back()->setPointSize(m_ui->pointSizeBox->value());
                m_cloudRenderer.back()->setRenderHeightColors(&m_renderHeightColors);
                m_ui->widget->addRenderable(item->data(1,Qt::DisplayRole).toString().toStdString(), [&,i](){ m_cloudRenderer[i]->render(); });
                m_ActivePointClouds.at(j) = true;
            }       
        }
    }

    // set origin-render-function
    for(unsigned int i = 0; i < m_PointClouds.size(); i++)
    {
    std::string identifier = m_PointClouds.at(i)->getLabelFile();
        identifier.push_back('_origin');
        m_ui->widget->addRenderable(identifier, [&,i](){

                                  if(m_ui->originsCheckBox->isChecked())
                                  {
                                      glPushMatrix();

                                      if(m_ActivePointClouds.at(i))
                                            glColor3f(1.0f,0.0f,0.0f);
                                      else
                                            glColor3f(0.5f,0.0f,0.0f);

                                      glTranslatef(
                                                   m_PointClouds.at(i)->getOrigin()(0),
                                                   m_PointClouds.at(i)->getOrigin()(1),
                                                   m_PointClouds.at(i)->getOrigin()(2));
                                      glutSolidSphere(0.5,10,10);

                                      glPopMatrix();
                                  }
                                });
    }
    m_ui->showPointsCheckBox->setCheckState(Qt::Checked); //overwrite gui button
}
//-----------------------------------------------------------------------------
void MainWindow::loadDataset()
{
    QString mapFile = QFileDialog::getOpenFileName(this, tr("Open Dataset"), "", tr("Dataset Files (*.js)"),0,QFileDialog::DontUseNativeDialog);
    if(mapFile=="") return;

    m_MapFile = mapFile;

    if(m_ui->scanTreeWidget->topLevelItemCount() > 0)
    {
        QMessageBox msgBox;
        msgBox.setText("Attempting to load new map.");
        msgBox.setInformativeText("Do you want to save eventual changes?");
        msgBox.setStandardButtons(QMessageBox::Save | QMessageBox::Discard | QMessageBox::Cancel);
        msgBox.setDefaultButton(QMessageBox::Save);
        switch (msgBox.exec()) {
           case QMessageBox::Save:    saveDataset();  m_ui->scanTreeWidget->clear(); m_PointClouds.clear(); m_ActivePointClouds.clear(); break;
           case QMessageBox::Discard: m_ui->scanTreeWidget->clear(); m_PointClouds.clear(); m_ActivePointClouds.clear(); break;
           case QMessageBox::Cancel:  return;
           default:                   std::cout << "An error occured in QMessageBox" << std::endl; break;
         }
    }

    std::string directory = QFileInfo(mapFile).absoluteDir().absolutePath().toStdString();

    //populate tree widget:
    boost::property_tree::ptree mapPT;
    boost::property_tree::json_parser::read_json(mapFile.toStdString(), mapPT);
    QList<QTreeWidgetItem*> items;
    BOOST_FOREACH(boost::property_tree::ptree::value_type& v, mapPT.get_child("scans"))
    {
        QTreeWidgetItem* entry = new QTreeWidgetItem(m_ui->scanTreeWidget);
        QString id =QString::fromStdString(v.second.get<std::string>("nodeId"));
        while(id.size() < 13)
            id.prepend('0'); //add starting zeros for automatic quicksort
        entry->setText(0,id);
        entry->setText(1,QString::fromStdString(directory) + "/" + QString::fromStdString(v.second.get<std::string>("nodeId")));
        items.append(entry);
    }
    m_ui->scanTreeWidget->insertTopLevelItems(0, items);

    QProgressDialog progress("Loading dataset from hard disk...", "Abort...", 0, items.size(), this, Qt::Dialog);
    progress.show(); //force show because saving is so fast, default is invisible for duration < 2 seconds

    BOOST_FOREACH(boost::property_tree::ptree::value_type& v, mapPT.get_child("scans"))
    {
        if (progress.wasCanceled())
        {
            std::cout << "Map loading aborted, removing all point clouds...";
            m_PointClouds.clear();
            m_ActivePointClouds.clear();
            std::cout << "\tdone." << std::endl;
            break;
        }

        std::string baseName = QFileInfo(QString::fromStdString(v.second.get<std::string>("nodeId"))).baseName().toStdString();
        std::string labelPath = directory + "/" + baseName + ".label";
        std::string veloDataPath = directory + "/" + baseName + ".velodata";

        if(!QFileInfo(QString::fromStdString(labelPath)).exists()) //only if no labels exists
        {

            std::ofstream newLabelFile;
            newLabelFile.open(labelPath);

            char veloBuffer[256];
            sprintf(veloBuffer,veloDataPath.c_str());
            FILE* veloFile=fopen(veloBuffer,"r");

            unsigned int r_size = 0;
            fread(&r_size,sizeof(unsigned int),1,veloFile); //read number of points

            for(unsigned int i=0; i < r_size; i++)
                newLabelFile << "Unknown\n";//fill file with Unknown label lines for each 3D point

            fclose(veloFile);
            newLabelFile.close();
            std::cout << "Created label file: " << labelPath  << " with " << r_size << " lines." << std::endl;
        }

        char veloBuffer[256];
        sprintf(veloBuffer,veloDataPath.c_str());
        FILE* veloFile=fopen(veloBuffer,"r");
        std::shared_ptr<PMap::PointCloud3d> pointCloud = std::make_shared<PMap::PointCloud3d>();

        unsigned int r_size = 0;
        fread(&r_size,sizeof(unsigned int),1,veloFile); //read number of points
        pointCloud->resize(r_size);

        unsigned int r_numTimeStamps = 0;
        fread(&r_numTimeStamps, sizeof(unsigned int), 1, veloFile); //read number of rawDataTimestamps
        std::vector<double> r_rawDataTimestamps;
        r_rawDataTimestamps.resize(r_numTimeStamps);
        fread(&r_rawDataTimestamps[0], sizeof(double), r_numTimeStamps, veloFile);//read rawDataTimestamps

        int r_circle_count = 0;
        fread(&r_circle_count,sizeof(int),1,veloFile); //read circle_count

        std::vector<float> tmp; //temp storage for all 3D measurements to create the pointCloud original cloud
        tmp.resize(r_size*3); // x,y,z foreach

        std::vector<float> r_Raw_circles; //temp storage for all raw_circles measurements
        r_Raw_circles.resize(r_size);

        std::vector<unsigned char> r_Inten_circles; //temp storage for all inten_circles measurements
        r_Inten_circles.resize(r_size);

        std::vector<Eigen::Vector3f> r_Circles; //temp storage for all circles measurements
        r_Circles.resize(r_size);

        std::ifstream labelFile;
        labelFile.open(labelPath);
        pointCloud->setLabelFile(labelPath);

        Eigen::Matrix3f R = Eigen::Matrix3f::Identity();
        Eigen::Vector3f t = Eigen::Vector3f::Zero();

        std::vector<float> tempValues;
        BOOST_FOREACH(boost::property_tree::ptree::value_type& w, v.second.get_child("pose"))
            tempValues.push_back(w.second.get<float>(""));

        Eigen::Quaternionf q(tempValues.at(3), tempValues.at(4), tempValues.at(5), tempValues.at(6));
        R = q.toRotationMatrix();
        t.x() = tempValues.at(0)*100; t.y() = tempValues.at(1)*100; t.z() = tempValues.at(2)*100; //! Warning: evil 100 scaling...

        int index = 0;
        for(int i=0; i<r_circle_count; i++)
        {
            unsigned int circle_size = 0;
            fread(&circle_size,sizeof(unsigned int),1,veloFile); //read #points per circle

            for(unsigned int j=0; j<circle_size; j++)
            {
                float raw_measurement = 0.0f;
                fread(&raw_measurement,sizeof(float),1,veloFile); //read all raw measurements
                r_Raw_circles.at(j) = raw_measurement;

                unsigned char inten_measurement = 0;
                fread(&inten_measurement,sizeof(unsigned char),1,veloFile); //read all intensity measurements
                r_Inten_circles.at(j) = inten_measurement;

                Eigen::Vector3f measurement3D(0.0f, 0.0f, 0.0f);
                fread(&measurement3D,sizeof(Eigen::Vector3f),1,veloFile); //read all 3D point measurements
                r_Circles.at(j) = measurement3D;

                //store as rotated and translated point in vector for the pointcloud
                measurement3D = R * measurement3D + t;
                tmp.at(index)   = measurement3D.x();
                tmp.at(index+1) = measurement3D.y();
                tmp.at(index+2) = measurement3D.z();

                std::string currentLabel;
                std::getline(labelFile, currentLabel);  //read label
                if(currentLabel ==    "Obstacle") pointCloud->setColor(index/3, OBSTACLE); //div by 3 because only 1 color for x,y,z value...
                else if(currentLabel == "Street") pointCloud->setColor(index/3, STREET); //R G B
                else if(currentLabel ==  "Rough") pointCloud->setColor(index/3, ROUGH);
                else                  /*Unknown*/ pointCloud->setColor(index/3, UNKNOWN);
                index+=3; //+3 foreach x,y,z point

            }
        }

        std::vector<Eigen::Vector3f> r_Points; //! temp storage for all calculated 3D points in it (measure-angle-ordered)
        r_Points.resize(r_size);
        std::vector<float> r_Raw_distances; //! temp storage for all distances measured (measure-angle-ordered)
        r_Raw_distances.resize(r_size);
        std::vector<unsigned char> r_Intensity; //! temp storage for intensities (measure-angle-ordered)
        r_Intensity.resize(r_size);

        for(unsigned int k = 0; k < r_size; k++)
        {
            Eigen::Vector3f point3D(0.0f, 0.0f, 0.0f);
            fread(&point3D,sizeof(Eigen::Vector3f),1,veloFile); //read all 3D points
            r_Points.at(k) = point3D;

            float raw_distance = 0.0f;
            fread(&raw_distance,sizeof(float),1,veloFile); //read all raw distances
            r_Raw_distances.at(k) = raw_distance;

            unsigned char intensityValue = '0';
            fread(&intensityValue,sizeof(unsigned char),1,veloFile); //read all intensity values
            r_Intensity.at(k) = intensityValue;
        }

        pointCloud->getPoints()=Eigen::Map<Eigen::Matrix<float,3,Eigen::Dynamic,Eigen::DontAlign> >(&tmp[0],3,r_size).cast<double>(); //optional, could use
        pointCloud->getOrigin() = t/100;

        pointCloud->setVertexArray();

        m_PointClouds.push_back(pointCloud);
        m_ActivePointClouds.push_back(false);

        labelFile.close();
        fclose(veloFile);

        progress.setValue(progress.value()+1);
    }

    progress.setValue(progress.maximum());
    if(items.first()) items.first()->setSelected(true); //draw at least 1 scan, calls loadVelodata()

}
//-----------------------------------------------------------------------------
void MainWindow::saveDataset()
{
    QProgressDialog progress("Saving all labels to hard disk...", "Abort save", 0, m_PointClouds.size(), this, Qt::Dialog);
    progress.show(); //force show because saving is so fast, default is invisible for duration < 2 seconds

    for(unsigned int i = 0; i < m_PointClouds.size(); i++)
    {
        if (progress.wasCanceled()) break;

        std::ofstream labelFile;
        labelFile.open(m_PointClouds.at(i)->getLabelFile(), std::ofstream::out | std::ofstream::trunc);//open & delete all contents

        for(unsigned int j = 0; j < m_PointClouds.at(i)->getPoints().cols(); j++)
        {
            if     (m_PointClouds.at(i)->getColor(j) == OBSTACLE) labelFile << "Obstacle\n";
            else if(m_PointClouds.at(i)->getColor(j) ==   STREET) labelFile << "Street\n";
            else if(m_PointClouds.at(i)->getColor(j) ==    ROUGH) labelFile << "Rough\n";
            else                                      /*Unknown*/ labelFile << "Unknown\n";
        }

        labelFile.close();
        progress.setValue(i);

    }
    progress.setValue(progress.maximum());
    sleep(1); //show QProgressDialog for at least a couple of seconds
}
//-----------------------------------------------------------------------------
void MainWindow::pointSizeChanged(int val)
{
    for (auto& r:m_cloudRenderer)
        r->setPointSize(val);
}
//-----------------------------------------------------------------------------
void MainWindow::labelChanged(QString label)
{
         if(label == "Obstacle") m_Annotation.setLabel(OBSTACLE);
    else if(label == "Rough")    m_Annotation.setLabel(ROUGH);
    else if(label == "Street")   m_Annotation.setLabel(STREET);
    else                         m_Annotation.setLabel(UNKNOWN);
}
//-----------------------------------------------------------------------------
void MainWindow::keyPressEvent(QKeyEvent* e)
{
    m_ui->widget->keyPressEvent(e);

    if(e->key() == Qt::Key_C)
        m_renderHeightColors = true;
}
//-----------------------------------------------------------------------------
void MainWindow::keyReleaseEvent(QKeyEvent* e)
{
    m_ui->widget->keyReleaseEvent(e);

    if(e->key() == Qt::Key_C)
        m_renderHeightColors = false;
}
//-----------------------------------------------------------------------------
void MainWindow::activatePointCloud(int i)
{

    bool activate = !m_ActivePointClouds.at(i);
    QString basename = QFileInfo(QString::fromStdString(m_PointClouds.at(i)->getLabelFile())).baseName() + ".velodata";

    m_ui->scanTreeWidget->findItems(basename, Qt::MatchContains).at(0)->setSelected(activate);
}
//-----------------------------------------------------------------------------
void MainWindow::showPointsChanged(int state)
{
    int j = 0;
    for (unsigned int i = 0; i < m_PointClouds.size(); i++)
        if(m_ActivePointClouds.at(i))
        {
            m_cloudRenderer.at(j)->setVisibility(state);
            j++;
        }
}
//-----------------------------------------------------------------------------
