/****************************************************************************
PointCloud.h
    is part of TeclMapVis

    Copyright 2014 Active Vision Group
              Universitaet Koblenz-Landau

    Author:  Marcel Haeselich
    Contact: Marcel Haeselich, mhaeselich[at]uni-koblenz.de


    TeclMapVis is free software: you can redistribute it and/or modify
    it under the terms of thr GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.
    TeclMapVis is distributed in the hope that it will be usefull,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public Livense for more details.
    You should have received a copy of the GNU General Public License
    along with TeclMapVis. If not, see <http://www.gnu.org/licenses/>.

****************************************************************************/

#ifndef PMAPUTIL_POINTCLOUD_H
#define PMAPUTIL_POINTCLOUD_H

#include <Eigen/Core>
#include <Eigen/StdVector>
#include <Eigen/Geometry>
#include <tuple>
#include <iostream>

namespace PMap
{
using namespace Eigen;

template<size_t Dim_>
class PointCloud
{
public:
    static_assert(Dim_==2||Dim_==3,"PointClouds only make sense in 2d or 3d space");

	static const int Dim=Dim_;

    Eigen::Vector3f& getOrigin() { return origin_; }
    const Eigen::Vector3f& getOrigin() const { return origin_; }

	Eigen::Matrix<double,Dim,Eigen::Dynamic,Eigen::DontAlign>& getPoints() { return points_; }
	const Eigen::Matrix<double,Dim,Eigen::Dynamic,Eigen::DontAlign>& getPoints() const { return points_; }

	Eigen::Matrix<double,Dim,Eigen::Dynamic,Eigen::DontAlign>& getNormals() { return normals_; }
	const Eigen::Matrix<double,Dim,Eigen::Dynamic,Eigen::DontAlign>& getNormals() const { return normals_; }

	std::vector<double>& getIntensities() { return intensities_; }
	const std::vector<double>& getIntensities() const { return intensities_; }

    void setColor(int index, std::tuple<float,float, float> color)
    {
        colorsArray_[index*3] = std::get<0>(color);
        colorsArray_[index*3+1] = std::get<1>(color);
        colorsArray_[index*3+2] = std::get<2>(color);
    }

    const std::tuple<float,float, float> getColor(int indx) const
    {
        return std::make_tuple(colorsArray_[3*indx], colorsArray_[3*indx+1], colorsArray_[3*indx+2]);
    }

	size_t getNumPoints() const { return static_cast<size_t>(points_.cols()); }

	// resizes the internal array of points to the given size -- all current points are lost
	void resize(size_t n);

	void conservativeResize(size_t n);

	bool hasNormals() const { return 0!=normals_.cols(); }
	void allocateNormals() { normals_.resize(Dim,points_.cols());}

	bool hasIntensities() const { return !intensities_.empty(); }
	void allocateIntensities() { intensities_.resize(points_.cols());}

    std::string getLabelFile() { return m_LabelFile;}
    void setLabelFile(std::string labelFile) { m_LabelFile = labelFile;}


    void allocateColorArray() {
        delete [] colorsArray_;
        colorsArray_ = new float[points_.cols()*3];
    }

    float* getColorArray() { return colorsArray_; }
    float* getHeightColorArray() { return heightColorArray_; }

    void setVertexArray(){
        delete [] vertexArray_;
        vertexArray_ = new float[points_.cols()*3];

        delete [] heightColorArray_;
        heightColorArray_ = new float[points_.cols()*3];

        for(int i = 0; i < points_.cols(); i++)
        {
            Eigen::Vector3d pt=points_.col(i);

            pt /= 100.0;
            float z = pt.z();


            float w = 3.0;

            while(z<0)z+=w;
            while(z>=w)z-=w;

            float r = 0.0;

            float b = z * 2.0 / w;
            if(b > 1.0)
                b = 2.0 - b;

            float g = 1.0 - b;

            vertexArray_[3*i] = pt.x();
            vertexArray_[3*i+1] = pt.y();
            vertexArray_[3*i+2] = pt.z();

            heightColorArray_[3*i] = b;
            heightColorArray_[3*i+1] = g;
            heightColorArray_[3*i+2] = r;
        }
    }

    float* getVertexArray() {return vertexArray_;}

private:
    Eigen::Vector3f origin_;
    Eigen::Matrix<double,Dim,Eigen::Dynamic,Eigen::DontAlign> points_;
    Eigen::Matrix<double,Dim,Eigen::Dynamic,Eigen::DontAlign> normals_;
    std::vector<double> intensities_;

    float* colorsArray_;
    float* vertexArray_;
    float* heightColorArray_;

    std::string m_LabelFile;
};

typedef PointCloud<2> PointCloud2d;
typedef PointCloud<3> PointCloud3d;

//-----------------------------------------------------------------------------
template<size_t Dim_>
void PointCloud<Dim_>::resize(size_t n)
{
	points_.resize(Dim,n);
	if (normals_.cols())
		normals_.resize(Dim,n);
	if (!intensities_.empty())
        intensities_.resize(n);

    allocateColorArray();

}
//-----------------------------------------------------------------------------
template<size_t Dim_>
void PointCloud<Dim_>::conservativeResize(size_t n)
{
	points_.conservativeResize(Dim,n);
	if (normals_.cols())
		normals_.conservativeResize(Dim,n);
	if (!intensities_.empty())
		intensities_.resize(n);

    allocateColorArray();
}
//-----------------------------------------------------------------------------
template<size_t Dim>
void transform(PointCloud<Dim>& p, const Matrix<double,Dim,Dim>& R, const Matrix<double,Dim,1>& t)
{
	p.getPoints()=(R*p.getPoints()).colwise()+t;
}
//-----------------------------------------------------------------------------
template<size_t Dim>
PointCloud<Dim> transformed(PointCloud<Dim> p, const Matrix<double,Dim,Dim>& R, const Matrix<double,Dim,1>& t)
{
	PointCloud<Dim> ret=std::move(p);
	transform(ret,R,t);
	return ret;
}
//-----------------------------------------------------------------------------
}

#endif
