#include "RenderWidget.h"
#include <GL/glew.h>
#include <GL/glut.h>
#include <iostream>
#include <boost/date_time/posix_time/posix_time_types.hpp>

namespace
{
void render_coordinateAxes(double size)
{
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glLineWidth(2.5);
	glColor3f(0, 1, 0);
	glBegin(GL_LINES);
	glVertex3f(0, 0, 0);
	glVertex3f(0, size, 0);
	glColor3f(1, 0, 0);
	glVertex3f(0, 0, 0);
	glVertex3f(size, 0, 0);
	glColor3f(0, 0, 1);
	glVertex3f(0, 0, 0);
	glVertex3f(0, 0, size);
	glEnd();
	glLineWidth(1);

	glDisable(GL_BLEND);
	glDisable(GL_LINE_SMOOTH);
}
double getCurrentTime()
{
	return boost::posix_time::microsec_clock::local_time().time_of_day().total_milliseconds();
}

}

//-----------------------------------------------------------------------------
RenderWidget::RenderWidget(QWidget* parent) : QGLWidget{QGLFormat{QGL::StencilBuffer | QGL::AlphaChannel},parent}, inputState{0}
{
    m_cam.setPos({2,0,1});
    m_cam.setLookAt({0,0,0});
    m_cam.setUp({0,0,1});

	QTimer* timer=new QTimer{this};
	connect(timer, SIGNAL(timeout()), this, SLOT(updateGL()));
	timer->start(1000/120);
	setMouseTracking(true);


    addRenderable("coordinateAxes",[](){
        render_coordinateAxes(1);
    });
}
//-----------------------------------------------------------------------------
void RenderWidget::initializeGL()
{
	glEnable(GL_DEPTH_TEST);
	glewInit();
}
//-----------------------------------------------------------------------------
void RenderWidget::handleInputs()
{
	// keyboard update
	{
		static double lastTime=getCurrentTime();
		const double curTime=getCurrentTime();
		const float dt=(curTime-lastTime)/1000.0f;
		lastTime=curTime;
		float speed=4.3*dt;
		if (inputState.spaceDown)
			speed*=5;

		if (inputState.forwardDown)
		{
			if (inputState.ctrlDown)
                m_cam.truckCamera(speed);
			else
                m_cam.moveCamera(speed);
		}
		else if (inputState.backwardDown)
		{
			if (inputState.ctrlDown)
                m_cam.truckCamera(-speed);
			else
                m_cam.moveCamera(-speed);
		}

		if (inputState.rightDown)
            m_cam.strafeCamera(speed);
		else if (inputState.leftDown)
            m_cam.strafeCamera(-speed);

	}

	// mouse update
    if ((inputState.leftMouseDown)&&(!inputState.ctrlDown))
	{
		QPoint local=mapFromGlobal(QCursor::pos());
		int x=local.x();
		int y=local.y();
		int dx=x-inputState.lastx;
		int dy=y-inputState.lasty;

        m_cam.setViewByMouse(dx, dy);

		inputState.lastx=x;
		inputState.lasty=y;
	}

}
//-----------------------------------------------------------------------------
void RenderWidget::paintGL()
{
	handleInputs();

	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
    gluLookAt(m_cam.getPos().x(), m_cam.getPos().y(), m_cam.getPos().z(),
              m_cam.getLookAt().x(), m_cam.getLookAt().y(), m_cam.getLookAt().z(),
              m_cam.getUp().x(), m_cam.getUp().y(), m_cam.getUp().z());

    render_coordinateAxes(1);

    m_Annotation->render_pen(this->mapFromGlobal( QCursor::pos() ));

    for (auto& p:m_renderables)
	{
		if (!p.second.enabled)
			continue;

		p.second.fn();
	}

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

	resizeGL(width(), height());
}
//-----------------------------------------------------------------------------
void RenderWidget::resizeGL(int w, int h)
{
	glViewport(0, 0, w, h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(85, float(w)/float(h), 0.1, 200.0);
	resolution.x=w;
	resolution.y=h;
}
//-----------------------------------------------------------------------------
void RenderWidget::keyPressEvent(QKeyEvent* e)
{
    if ((e->key()==Qt::Key_W) || (e->key()==Qt::Key_Up))
		inputState.forwardDown=true;
	else if ((e->key()==Qt::Key_S) || (e->key()==Qt::Key_Down))
		inputState.backwardDown=true;
	else if ((e->key()==Qt::Key_A) || (e->key()==Qt::Key_Left))
		inputState.leftDown=true;
	else if ((e->key()==Qt::Key_D) || (e->key()==Qt::Key_Right))
		inputState.rightDown=true;

	if (e->key()==Qt::Key_Space)
		inputState.spaceDown=true;

	if (e->key()==Qt::Key_Control)
		inputState.ctrlDown=true;

    m_Annotation->keyPressEvent(e);
}
//-----------------------------------------------------------------------------
void RenderWidget::keyReleaseEvent(QKeyEvent* e)
{
	if ((e->key()==Qt::Key_W) || (e->key()==Qt::Key_Up))
		inputState.forwardDown=false;
	else if ((e->key()==Qt::Key_S) || (e->key()==Qt::Key_Down))
		inputState.backwardDown=false;
	else if ((e->key()==Qt::Key_A) || (e->key()==Qt::Key_Left))
		inputState.leftDown=false;
	else if ((e->key()==Qt::Key_D) || (e->key()==Qt::Key_Right))
		inputState.rightDown=false;

	if (e->key()==Qt::Key_Space)
		inputState.spaceDown=false;

	if (e->key()==Qt::Key_Control)
		inputState.ctrlDown=false;

    m_Annotation->keyReleaseEvent(e);
}
//-----------------------------------------------------------------------------
void RenderWidget::mousePressEvent(QMouseEvent* e)
{
	if (e->button()==Qt::LeftButton)
	{
		inputState.leftMouseDown=true;

		QPoint local=mapFromGlobal(QCursor::pos());

		inputState.lastx=local.x();
		inputState.lasty=local.y();
	}
	else if (e->button()==Qt::RightButton)
	{
		inputState.rightMouseDown=true;
	}

    m_Annotation->mousePressEvent(e);
}
//-----------------------------------------------------------------------------
void RenderWidget::mouseDoubleClickEvent(QMouseEvent* e)
{
    m_Annotation->mouseDoubleClickEvent(e);
}
//-----------------------------------------------------------------------------
void RenderWidget::wheelEvent(QWheelEvent* e)
{
    m_Annotation->wheelEvent(e);
}
//-----------------------------------------------------------------------------
void RenderWidget::mouseReleaseEvent(QMouseEvent* e)
{
	if (e->button()==Qt::LeftButton)
    {
		inputState.leftMouseDown=false;
	}
	else if (e->button()==Qt::RightButton)
    {
		inputState.rightMouseDown=false;
    }

    m_Annotation->mouseReleaseEvent(e);
}
//-----------------------------------------------------------------------------
void RenderWidget::mouseMoveEvent(QMouseEvent* e)
{
    m_Annotation->mouseMoveEvent(e);
}
//-----------------------------------------------------------------------------
void RenderWidget::addRenderable(const std::string& name, std::function<void()> fn)
{
    m_renderables.insert(std::make_pair(name,Renderable{name,true,fn}));
}
//-----------------------------------------------------------------------------
void RenderWidget::removeRenderable(const std::string& name)
{
    m_renderables.erase(name);
}
//-----------------------------------------------------------------------------
void RenderWidget::enableRenderable(const std::string& name, bool enabled)
{
    auto it=m_renderables.find(name);
    if (it!=m_renderables.end())
		it->second.enabled=enabled;
}
//-----------------------------------------------------------------------------
void RenderWidget::clearRenderables()
{
    m_renderables.clear();
}
//-----------------------------------------------------------------------------
